using System;

namespace dipl_rad_1589
{	
	/// <summary>
	/// Klasa Solver sadrzi sve metode i podatke potrebne za simulaciju.
	/// </summary>
	public class Solver
	{
		// Sirina i visina mreze elemenata simuliranoga sustava
		public int WIDTH, HEIGHT;

		// Brzina u tekucem i prethodnom koraku
		public double[,,] vel;
		private double[,,] oldVel;

		// Trenutni tlak
		public double[,] p;

		// Korekcije tlaka u tekucem i prethodnom koraku
		private double[,] pCorrection;
		private double[,] oldPCorrection;

		// Mapa prepreka
		public int[,] obstacleMap;

		// Parametri simulacije
		private double maxPCorrection;
		public double deltaT, deltaX, deltaY;
		public double Re;
		public double convergenceCondition = 0.00001;
		public int maxIteracija = 100;

		public double velUndRelax = 0.8;
		public double pUndRel = 0.5;


		/// <summary>
		/// Konstruktor klase. Instancira sva polja i postavlja defaultne vrijednosti parametara simulacije.
		/// </summary>
		/// <param name="WIDTH">Sirina simuliranoga sustava</param>
		/// <param name="HEIGHT">Visina simuliranoga sustava</param>
		public Solver(int WIDTH, int HEIGHT)
		{
			// Postavi lokalne dimenzije
			this.WIDTH = WIDTH;
			this.HEIGHT = HEIGHT;

			// Instanciraj objekte za potrebna polja
			vel = new double[WIDTH + 1, HEIGHT + 1, 2];
			oldVel = new double[WIDTH + 1, HEIGHT + 1, 2];

			p = new double[WIDTH, HEIGHT];
			pCorrection = new double[WIDTH, HEIGHT];
			oldPCorrection = new double[WIDTH, HEIGHT];

			obstacleMap = new int[WIDTH, HEIGHT];

			// Postavi pocetne vrijednosti parametara
			deltaT = 0.01;
			deltaX = 1;
			deltaY = 1;

			Re = 400;

			// Postavi pocetne vrijednosti polja brzine, tlaka i prepreka na nula
			setZeroInitialVelocities();
			setZeroInitialPressure();
			setNoInitialObstacles();
		}

		// Postavi pocetne vrijednosti polja brzina na nula
		private void setZeroInitialVelocities () 
		{
			for (int i = 0; i < vel.GetLength(0); i++) 
			{
				for (int j = 0; j < vel.GetLength(1); j++) 
				{
					vel[i, j, 0] = 0;
					vel[i, j, 1] = 0;

					oldVel[i, j, 0] = 0;
					oldVel[i, j, 1] = 0;
				}
			}
		}

		// Postavi pocetne vrijednosti polja tlaka na nula
		private void setZeroInitialPressure () 
		{
			for (int i = 0; i < p.GetLength(0); i++) 
			{
				for (int j = 0; j < p.GetLength(1); j++) 
				{
					p[i, j] = 0;
					pCorrection[i, j] = 0;
				}
			}
		}

		// Postavi pocetne vrijednosti polja prepreka na nula
		private void setNoInitialObstacles () 
		{
			for (int i = 0; i < obstacleMap.GetLength(0); i++) 
			{
				for (int j = 0; j < obstacleMap.GetLength(1); j++) 
				{
					obstacleMap[i, j] = 0;
				}
			}
		}

		// Prihvati novi raspored prepreka u sustavu
		public void setObstacles (int[,] obstacleMap) 
		{
			for (int i = 0; i < this.obstacleMap.GetLength(0); i++) 
			{
				for (int j = 0; j < this.obstacleMap.GetLength(1); j++) 
				{
					this.obstacleMap[i, j] = obstacleMap[i, j];
				}
			}
		}

		/// <summary>
		/// Sredisnji dio klase i procesa simulacije
		/// </summary>
		/// <returns>true ako je do konvergencije doslo, false inace</returns>
		public bool advanceTimeStep () 
		{
			// Zamijeni oldVel i vel vrijednosti
			switchVelocities ();
			
			// Ponavljaj postupak do konvergencije ili najvise maxIteracija puta
			for (int iteracija = 0; iteracija < maxIteracija; iteracija++) 
			{
				// Izracunaj nove iznose brzina koristeci momentnu jednadzbu (NS jednadzbe)
				calculateNewVelocities ();

				// Izracunaj koreksije tlaka koristeci Poissonovu jednadzbu
				calculatePCorrection ();

				// Primijeni korekcije tlaka
				applyPCorrection ();

				// Ako je najveca korekcija tlaka dovoljno mala, smatra se da je postupak dosao do konvergencije
				if (maxPCorrection < convergenceCondition) 
				{
					// Primijeni korekcije brzina
					applyVelocityCorrections ();

					return true;
				}
			}
			return false;
		}

		// Zamijeni polja starih i novih brzina. Zbog ubrzanja postupka mijenjaju se reference da bi se izbjeglo kopiranje kompletnog sadrzaja
		private void switchVelocities () 
		{
			double[,,] tempVel;

			tempVel = vel;
			vel = oldVel;
			oldVel = tempVel;
		}

		// Postavi granice u ovisnosti o mapi prepreka
		public void setBoundaries () 
		{
			// Za svaki element mreze
			for (int i = 0; i < WIDTH; i++) 
			{
				for (int j = 0; j < HEIGHT; j++) 
				{
					// Ako je element utok postavi brzine ovisno na kojoj se strani nalazi
					if (obstacleMap[i, j] == 2) 
					{
						if (i == 0)
						{
							vel[i, j, 0] = 0.5;
							continue;
						}
						if (i == WIDTH - 1) 
						{
							vel[i, j, 0] = -0.5;
							continue;
						}
						if (j == 0)
						{
							vel[i, j, 1] = 0.5;
							continue;
						}
						if (j == HEIGHT - 1) 
						{
							vel[i, j, 1] = -0.5;
							continue;
						}

						// Ako se element ne nalazi na rubu, ne moze biti utok.
						obstacleMap[i, j] = 0;
					}
					// Ako je element istok postavi brzine ovisno na kojoj se strani nalazi
					if (obstacleMap[i, j] == 3) 
					{
						if (i == 0)
						{
							vel[i, j, 0] = -0.5;
							continue;
						}
						if (i == WIDTH - 1) 
						{
							vel[i, j, 0] = 0.5;
							continue;
						}
						if (j == 0)
						{
							vel[i, j, 1] = -0.5;
							continue;
						}
						if (j == HEIGHT - 1) 
						{
							vel[i, j, 1] = 0.5;
							continue;
						}
						// Ako se ne nalazi na rub, ne moze biti istok
						obstacleMap[i, j] = 0;
					}
					// Ako je element pomicni zid, postavi brzine ovisno na kojoj se strani nalazi
					if (obstacleMap[i, j] == 4) 
					{
						if (i == 0)
						{
							vel[i, j, 1] = -0.5;
							continue;
						}
						if (i == WIDTH - 1) 
						{
							vel[i, j, 1] = 0.5;
							continue;
						}
						if (j == 0)
						{
							vel[i, j, 0] = -0.5;
							continue;
						}
						if (j == HEIGHT - 1) 
						{
							vel[i, j, 0] = 0.5;
							continue;
						}
						// Ako se ne nalazi na rubu, ne moze biti pomicni zid (u ovoj implementaciji)
						obstacleMap[i, j] = 0;
					}
				}
			}
		}

		// Za svaki element poziva metodu za izracun novih brzina
		private void calculateNewVelocities () 
		{
			for (int i = 0; i < vel.GetLength(0) - 1; i++) 
			{
				for (int j = 0; j < vel.GetLength(1) - 1; j++) 
				{
					if (obstacleMap[i, j] == 0)
						calculateOneElementNewVelocity(i, j);
				}
			}
		}

		/// <summary>
		/// Proracun nove brzine za zadani element koristenjem momentne jednadzbe
		/// </summary>
		/// <param name="i">Indeks i trazenog elementa</param>
		/// <param name="j">Indeks j trazenog elementa</param>
		private void calculateOneElementNewVelocity (int i, int j) 
		{
			// Zbog bolje citljivosti i jednostavnije implementacije za susjedne elemente se koriste varijable imena prema stranama svijeta
			double northPressure, southPressure, eastPressure, westPressure;
			double eastXVel, westXVel, farEastXVel, eastNorthXVel, eastSouthXVel;
			double northYVel, southYVel, farNorthYVel, northEastYVel, northWestYVel;

			// Postavi inicijalne vrijednosti bez provjere vrste elementa koji je susjed
			northPressure = p[i, j + 1];
			southPressure = p[i, j - 1];
			eastPressure = p[i + 1, j];
			westPressure = p[i - 1, j];

			eastXVel = oldVel[i, j, 0];
			westXVel = oldVel[i - 1, j, 0];
			farEastXVel = oldVel[i + 1, j, 0];
			eastNorthXVel = oldVel[i, j + 1, 0];
			eastSouthXVel = oldVel[i, j - 1, 0];

			northYVel = oldVel[i, j, 1];
			southYVel = oldVel[i, j - 1, 1];
			farNorthYVel = oldVel[i, j + 1, 1];
			northEastYVel = oldVel[i + 1, j, 1];
			northWestYVel = oldVel[i - 1, j, 1];

			// Provjeri vrstu susjeda i prilagodi susjedne varijable
			if (obstacleMap[i + 1, j] == 1) 
			{
				eastPressure = p[i, j];
				eastXVel = 0;
				farEastXVel = 0;
				northEastYVel = 0;
			}
			if (obstacleMap[i - 1, j] == 1) 
			{
				westPressure = p[i, j];
				westXVel = 0;
				northWestYVel = 0;
			}
			if (obstacleMap[i, j + 1] == 1) 
			{
				northPressure = p[i, j];
				northYVel = 0;
				farNorthYVel = 0;
				eastNorthXVel = 0;
			}
			if (obstacleMap[i, j - 1] == 1) 
			{
				southPressure = p[i, j];
				southYVel = 0;
				eastSouthXVel = 0;
			}
			if (obstacleMap[i - 1, j - 1] == 1) 
			{
			}
			if (obstacleMap[i - 1, j + 1] == 1) 
			{
				northWestYVel = 0;
			}
			if (obstacleMap[i + 1, j - 1] == 1) 
			{
				eastSouthXVel = 0;
			}
			if (obstacleMap[i + 1, j + 1] == 1) 
			{
				eastNorthXVel = 0;
				northEastYVel = 0;
			}
			if ((i + 2) < obstacleMap.GetLength(0)) 
			{
				if (obstacleMap[i + 2, j] == 1) 
				{
					farEastXVel = 0;
				}
			}
			else 
			{
				if (obstacleMap[i + 1, j] == 1) 
				{
					farEastXVel = 0;
				}
				else 
				{
					// Ovaj dio ce se izvrsiti za slucaj inflow/outflow na East strani
					farEastXVel = eastXVel;
				}
			}
			if ((j + 2) < obstacleMap.GetLength(1)) 
			{
				if (obstacleMap[i, j + 2] == 1) 
				{
					farNorthYVel = 0;
				}
			}
			else 
			{
				if (obstacleMap[i, j + 1] == 1) 
				{
					farNorthYVel = 0;
				}
				else 
				{
					// Ovaj dio ce se izvrsiti za slucaj inflow/outflow na North strani
					farNorthYVel = northYVel;
				}
			}

			// Primijeni postavljene varijable u momentnoj jednadzbi
			vel[i, j, 0] = oldVel[i, j, 0] - velUndRelax * deltaT * ((eastXVel * (eastXVel - westXVel) / deltaX + northYVel * (eastNorthXVel - eastXVel) / deltaY) + (eastPressure - p[i, j]) / deltaX - 1 / Re * ((farEastXVel - 2 * eastXVel + westXVel) / (deltaX * deltaX) + (eastNorthXVel - 2 * eastXVel + eastSouthXVel) / (deltaY * deltaY)));
			vel[i, j, 1] = oldVel[i, j, 1] - velUndRelax * deltaT * ((eastXVel * (northEastYVel - northYVel) / deltaX + northYVel * (northYVel - southYVel) / deltaY) + (northPressure - p[i, j]) / deltaY - 1 / Re * ((northEastYVel - 2 * northYVel + northWestYVel) / (deltaX * deltaX) + (farNorthYVel - 2 * northYVel + southYVel) / (deltaY * deltaY)));
		}

		// Za svaki element primijeni korekcije tlaka
		private void applyPCorrection () 
		{
			for (int i = 1; i < p.GetLength(0) - 1; i++) 
			{
				for (int j = 1; j < p.GetLength(1) - 1; j++) 
				{
					p[i, j] += pUndRel * pCorrection[i, j];
				}
			}
		}

		// Za svaki element primijeni korekcije brzina
		private void applyVelocityCorrections () 
		{
			for (int i = 1; i < vel.GetLength(0) - 2; i++) 
			{
				for (int j = 1; j < vel.GetLength(1) - 2; j++) 
				{
					vel[i, j, 0] += - deltaT / deltaX * (pCorrection[i + 1, j] - pCorrection[i, j]);
					vel[i, j, 1] += - deltaT / deltaY * (pCorrection[i, j + 1] - pCorrection[i, j]);
				}
			}
		}
	
		// Iterativni postupak rjesavanja Poissonove jednadzbe
		private bool calculatePCorrection () 
		{
			double maxDiff = 0;

			// Postavi elemente pola korekcija tlaka na nula
			for (int i = 0; i < WIDTH; i++) 
			{
				for (int j = 0; j < HEIGHT; j++) 
				{
					pCorrection[i, j] = 0;
					oldPCorrection[i, j] = 0;
				}
			}

			// Ponavljaj postupak do konvergencije ili najvise maxIteracija puta
			for (int numOfIterations = 0; numOfIterations < maxIteracija; numOfIterations++) 
			{
				maxDiff = 0;
				maxPCorrection = 0;
				
				for (int i = 1; i < p.GetLength(0) - 1; i++) 
				{
					for (int j = 1; j < p.GetLength(1) - 1; j++) 
					{
						// Ako se ne radi o elementu ispunjenom fluidom, nema potrebe za racunanjem
						if (obstacleMap[i, j] != 0)
							continue;

						// Izracunaj korekciju tlaka za tekuci element
						calculateOneElementNewPressure(i, j);
					
						// Zapamti najvecu apsolutnu promjenu korekcije tlaka
						if (Math.Abs(pCorrection[i, j] - oldPCorrection[i, j]) > maxDiff)
							maxDiff = Math.Abs(pCorrection[i, j] - oldPCorrection[i, j]);

						// Zapamti najvecu apsolutnu korekciju tlaka
						if (Math.Abs(pCorrection[i, j]) > maxPCorrection)
							maxPCorrection = Math.Abs(pCorrection[i, j]);
					}
				}

				// Ako je najveca promjena korekcije tlaka manja od uvjeta konvergencije, postupak se prekida
				if (maxDiff < convergenceCondition)
					return true;
				// inace tekuce korekcije postaju stare i postupak se nastavlja
				else 
				{
					for (int i = 0; i < p.GetLength(0); i++) 
					{
						for (int j = 0; j < p.GetLength(1); j++) 
						{
							oldPCorrection[i, j] = pCorrection[i, j];
						}
					}
				}
			}
			return false;
		}

		// Izracunaj korekciju tlaka za zadani element
		private void calculateOneElementNewPressure (int i, int j) 
		{
			// Koriste se varijable nazvane po stranama svijeta za vecu citljivost koda i jednostavniju implementaciju
			double northPressure, southPressure, eastPressure, westPressure;
			double eastXVel, westXVel;
			double northYVel, southYVel;

			// Postavi inicijalne vrijednosti bez obzira na vrstu susjednih elemenata
			northPressure = pCorrection[i, j + 1];
			southPressure = pCorrection[i, j - 1];
			eastPressure = pCorrection[i + 1, j];
			westPressure = pCorrection[i - 1, j];

			eastXVel = vel[i, j, 0];
			westXVel = vel[i - 1, j, 0];
			northYVel = vel[i, j, 1];
			southYVel = vel[i, j - 1, 1];

			// Prilagodi vrijednosti vrsti susjednih elemenata
			if (obstacleMap[i + 1, j] == 1) 
			{
				eastPressure = 0;
				eastXVel = 0;
			}
			if (obstacleMap[i - 1, j] == 1) 
			{
				westPressure = 0;
				westXVel = 0;
			}
			if (obstacleMap[i, j + 1] == 1) 
			{
				northPressure = 0;
				northYVel = 0;
			}
			if (obstacleMap[i, j - 1] == 1) 
			{
				southPressure = 0;
				southYVel = 0;
			}

            // Izracunaj korekcije tlaka koristeci Poissonovu jednadzbu			
			pCorrection[i, j] = (eastPressure + westPressure) * deltaT / (deltaX * deltaX) + (northPressure + southPressure) * deltaT / (deltaY * deltaY)  - (eastXVel - westXVel) / deltaX - (northYVel - southYVel) / deltaY;
			pCorrection[i, j] /= 2 * deltaT * (deltaX * deltaX + deltaY * deltaY) / (deltaX * deltaX * deltaY * deltaY);
		}

	}
}