using System;

namespace dipl_rad_1589
{
	/// <summary>
	/// Klasa Visualizer sadrzi metode potrebne za iscrtavanje rezultata Solver-a u zadani pictureBox. Skaliranje
	/// se vrsi automatski.
	/// </summary>
	public class Visualizer
	{
		// Objekti tipa Graphics, Pen i Brush su potrebni za iscrtavanja
		private System.Drawing.Graphics gr;
		private System.Drawing.Pen p = new System.Drawing.Pen(System.Drawing.Color.White);
		private System.Drawing.SolidBrush b = new System.Drawing.SolidBrush(System.Drawing.Color.Gray);

		// Referenca na objekt tipa solv. Postavlja se u konstruktoru.
		private Solver solv;

		// Objekt tipa Image, omogucava iscrtavanje neovisno o pictureBox-u sto olaksava refresh i spremanje podataka
		public System.Drawing.Bitmap slika;

		// Odnos velicine podrucja iscrtavanja i simuliranoga sustava. Za iznose manje od 2 iscrtavanje moze dati lose rezultate.
		public double zoom;

		// Odmaci podrucja iscrtavanja od granica pictureBoxa uzrokovani razlicitim odnosima stranica. Potrebni da
		// bi se slika iscrtavala centrirano.
		public int offsetX, offsetY;

		// Mapa boja. Koristi se za iscrtavanje tlaka i brzina. Sadrzi rezultate vizualizacije za vrijeme racunanja
		private System.Drawing.Color[,] colorMap;

		// Mapa prepreka. Sadrzi podatke o sustavu. Odreduje prepreke, utoke, istoke i pomicne zidove.
		private int[,] obstacleMap = new int[100, 100];
		
		// Sirina i visina simuliranoga sustava (mreze simulacije)
		public int problemWidth, problemHeight;

		/// <summary>
		/// Konstruktor klase. Instancira potrebne objekte i inicijalizira potrebne parametre.
		/// </summary>
		/// <param name="picBox">Objekt koji implementira createGraphics() i na koji ce se iscrtavati rezultati (ovdje iskljucivo pictureBox</param>
		/// <param name="problemWidth">Sirina simulacijske mreze</param>
		/// <param name="problemHeight">Visina simulacijske mreze</param>
		/// <param name="solv">Objekt tipa Solver koji sadrzi sve podatke vezane za simulaciju.</param>
		public Visualizer(System.Windows.Forms.PictureBox picBox, int problemWidth, int problemHeight, Solver solv)
		{
			// Instanciraj objekt tipa Image
			slika = new System.Drawing.Bitmap(picBox.Width, picBox.Height);

			// Image dobivenog pictureBoxa postavi na 'slika'
			picBox.Image = slika;

			// Instanciraj novi objekt tipa Graphics i ppovezi ga sa 'slika'
			gr = System.Drawing.Graphics.FromImage(slika);

			// Povezi lokalne varijable s dobivenima
			this.solv = solv;
			this.problemWidth = problemWidth;
			this.problemHeight = problemHeight;

			// Instanciraj polje colorMap potrebnih dimanzija
			colorMap = new System.Drawing.Color[problemWidth, problemHeight];

			// Izracunaj vrijednosti varijabli zoom, offsetX i offsetY tako da se podrucje iscrtavanja maksimalno
			// iskoristi, ali be rastezanja slike
			if ((double)picBox.Width/problemWidth > picBox.Height/problemHeight) 
			{
				zoom = (double)picBox.Height / problemHeight;
				offsetX = (int)((picBox.Width - problemWidth * zoom) / 2);
				offsetY = 0;
			}
			else 
			{
				zoom = (double)picBox.Width / problemWidth;
				offsetX = 0;
				offsetY = (int)((picBox.Height - problemHeight * zoom) / 2);
			}
		}

		/// <summary>
		/// Metoda koja obraduje promjenu velicine podrucja iscrtavanja. Potrebno je instancirati nove objekte zbog
		/// promijenjenih zahtjeva za memorijom.
		/// </summary>
		/// <param name="picBox">PictureBox na koji se iscrtava</param>
		public void canvasSizeChange (System.Windows.Forms.PictureBox picBox) 
		{
			// Unisti trenutni objekt
			slika.Dispose();

			// Instanciraj novi s promijenjenim dimenzijama i ponovno ga povezi s pictureBox-om i Graphics-om
			slika = new System.Drawing.Bitmap(picBox.Width, picBox.Height);
			picBox.Image = slika;
			gr = System.Drawing.Graphics.FromImage(slika);

			// Izracunaj nove vrijednosti varijabli zoom, offsetX i offsetY
			if ((double)picBox.Width/problemWidth > picBox.Height/problemHeight) 
			{
				zoom = (double)picBox.Height / problemHeight;
				offsetX = (int)((picBox.Width - problemWidth * zoom) / 2);
				offsetY = 0;
			}
			else 
			{
				zoom = (double)picBox.Width / problemWidth;
				offsetX = 0;
				offsetY = (int)((picBox.Height - problemHeight * zoom) / 2);
			}
		}

		/// <summary>
		/// Iscrtavanje mreze. Koristi se pri ucrtavanju prepreka i pri iscrtavanju tlaka i brzina (ako je tako odabrano)
		/// </summary>
		public void drawGrid () 
		{
			for (int i = 0; i <= problemWidth; i++) 
			{
				gr.DrawLine(p, (float)(offsetX + i * zoom), (float)(offsetY), (float)(offsetX + i * zoom), (float)(offsetY + problemHeight * zoom));
			}
			for (int i = 0; i <= problemHeight; i++) 
			{
				gr.DrawLine(p, (float)(offsetX), (float)(offsetY + i * zoom), (float)(offsetX + problemWidth * zoom), (float)(offsetY + i * zoom));
			}
		}

		// Pojednostavljeni oblik metode za lokalnu primjenu
		public void obstacleVisualization () 
		{
			obstacleVisualization (obstacleMap, false);
		}

		/// <summary>
		/// Iscrtavanje prepreka
		/// </summary>
		/// <param name="obstacleMap">Mapa prepreka koje treba iscrtati</param>
		/// <param name="initiating">Logicka vrijednost: true brise postojecu sliku i iscrtava samo prepreke, false dodaje prepreke na postojecu sliku</param>
		public void obstacleVisualization (int[,] obstacleMap, bool initiating) 
		{
			this.obstacleMap = obstacleMap;

			if (initiating)
				gr.Clear(System.Drawing.Color.LightGray);

			for (int i = 0; i < problemWidth; i++) 
			{
				for (int j = 0; j < problemHeight; j++) 
				{
					// Ako treba samo dodati prepreke, ne mijenjaj fluid
					if (obstacleMap[i, j] == 0 && !initiating)
						continue;
					// Ako treba iscrtati samo prepreke, prebrisi i fluidne elemente
					if (obstacleMap[i, j] == 0 && initiating) 
						b.Color = System.Drawing.Color.LightGray;
					// Prepreke - siva boja
					if (obstacleMap[i, j] == 1) 
						b.Color = System.Drawing.Color.Gray;
					// Utok - narancasta boja
					if (obstacleMap[i, j] == 2) 
						b.Color = System.Drawing.Color.Orange;
					// Istok - ljubicasta boja
					if (obstacleMap[i, j] == 3) 
						b.Color = System.Drawing.Color.Purple;
					// Pomicni zid - plava boja
					if (obstacleMap[i, j] == 4) 
						b.Color = System.Drawing.Color.Navy;

					gr.FillRectangle (b, (float)(offsetX + i * zoom), (float)(offsetY + j * zoom), (float)zoom, (float)zoom);
				}
			}

			if (initiating)
				drawGrid();
		}

		public void clearPicBox (System.Drawing.Color color) 
		{
			gr.Clear(color);
		}

		/// <summary>
		/// Racunanje srednje vrijednosti brzina za potrebe vizualizacije
		/// </summary>
		/// <param name="coor">Trazeni smjer: 0 - x, 1 - y</param>
		/// <returns>Srednja vrijednost brzine</returns>
		private double getMeanVelValue (int coor) 
		{
			double max = 0;

			// Trazi najvecu apsolutnu vrijednost brzine
			for (int i = 0; i < problemWidth + 1; i++) 
			{
				for (int j = 0; j < problemHeight + 1; j++) 
				{
					if (max < Math.Abs(solv.vel[i, j, coor]))
						max = Math.Abs(solv.vel[i, j, coor]);
				}
			}

			// Vrati polovinu najveceg iznosa
			return max/2;
		}

		// Srednja vrijednost tlaka za potrebe vizualizacije
		private double getMeanPValue () 
		{
			double sum = 0;

			// Zbroji sve apsolutne vrijednosti tlaka
			for (int i = 0; i < problemWidth; i++) 
			{
				for (int j = 0; j < problemHeight; j++) 
				{
					sum += Math.Abs(solv.p[i, j]);
				}
			}
			// Vrati aritmeticku sredinu 
			return sum / (problemHeight * problemWidth);
		}

		// Iscrtavanje stanja mape boja
		public void drawFromColorMap() 
		{			
			for (int i = 0; i < problemWidth; i++) 
			{
				for (int j = 0; j < problemHeight; j++) 
				{
					b.Color = colorMap[i, j];
					gr.FillRectangle(b, (float)(offsetX + i * zoom), (float)(offsetY + j * zoom), (float)(zoom), (float)(zoom));
				}
			}
			// Dodaj prepreke
			obstacleVisualization();
		}

		/// <summary>
		/// Iscrtavanje rezultata simulacije - vizualizacija iznosa brzina
		/// </summary>
		/// <param name="xVel">Logicka vrijednost - true: X komponenta se iscrtava, false: X komponenta se ne iscrtava</param>
		/// <param name="yVel">Logicka vrijednost - true: Y komponenta se iscrtava, false: Y komponenta se ne iscrtava</param>
		public void drawVelocities (bool xVel, bool yVel) 
		{
			double meanXValue = 1;
			double meanYValue = 1; 
			double meanVelocity = 1;
			int temp;
			
			// Izracunaj srednje vrijednosti brzina
			if (xVel) meanXValue = getMeanVelValue(0) + 0.001;		// 0.001 zbog izbjegavanja dijeljenja s nulom
			if (yVel) meanYValue = getMeanVelValue(1) + 0.001;

			// Izracunaj ukupnu srednju brzinu kao aritmeticku sredinu komponenti
			meanVelocity = (meanXValue + meanYValue) / 2;

			for (int i = 0; i < problemWidth; i++) 
			{
				for (int j = 0; j < problemHeight; j++ ) 
				{
					if (xVel) 
					{
						// Skaliraj iznos na 0-255 u ovisnosti o srednjoj vrijednosti
						temp = (int) Math.Abs((127 * ((solv.vel[i, j, 0] + solv.vel[i, j + 1, 0]) / (2 * meanVelocity))));
						colorMap[i, j] = System.Drawing.Color.FromArgb((temp > 255)? 255:temp, 0, 0);
					}
					else 
					{
						colorMap[i, j] = System.Drawing.Color.FromArgb(0, 0, 0);
					}

					if (yVel) 
					{
						// Skaliraj iznos na 0-255 u ovisnosti o srednjoj vrijednosti
						temp = (int) Math.Abs((127 * ((solv.vel[i, j, 1] + solv.vel[i + 1, j, 1]) / (2 * meanVelocity))));
						colorMap[i, j] = System.Drawing.Color.FromArgb(colorMap[i,j].R, (temp > 255)? 255: temp, 0);
					}
					else
					{
						colorMap[i, j] = System.Drawing.Color.FromArgb(colorMap[i,j].R, 0, 0);
					}
				}
			}
			// Iscrtaj dobivene rezultate
			drawFromColorMap();
		}

		/// <summary>
		/// Iscrtavanje rezultata simulacije - vizualizacija tlaka
		/// </summary>
		public void drawPressure () 
		{
			double meanPValue = 1;
			int temp;
			
			// Izracunaj srednju vrijednost tlaka
			meanPValue = getMeanPValue() + 0.001;		// 0.001 zbog izbjegavanja dijeljenja s nulom

			for (int i = 0; i < problemWidth; i++) 
			{
				for (int j = 0; j < problemHeight; j++ ) 
				{
					// Skaliraj iznos na 0-255 u ovisnosti o srednjoj vrijednosti
					temp = (int) Math.Abs((127 * (solv.p[i, j] / meanPValue)));
					colorMap[i, j] = System.Drawing.Color.FromArgb(0, 0, (temp > 255)? 255:temp);
				}
			}
			// Iscrtaj dobivene rezultate
			drawFromColorMap();
		}

		/// <summary>
		/// Iscrtavanje rezultata simulacije - vizualizacija vektora brzina
		/// </summary>
		public void drawStreamlines () 
		{
			double zoomDivider = 1;
			double meanVelocity = (getMeanVelValue(0) + getMeanVelValue(1)) / 2;

			clearPicBox(System.Drawing.Color.Black);

			for (int i = 0; i < problemWidth; i++) 
			{
				for (int j = 0; j < problemHeight; j++) 
				{
					// Iscrtaj dvije linije. Jedna crvena, jedna bijela, tako da se nastavljaju jedna na drugu 
					// i pokazuju smjer gibanja (od crvene prema bijeloj)
					p.Color = System.Drawing.Color.Red;
					gr.DrawLine(p, (float)(offsetX + i * zoom - 0.5 * solv.vel[i, j, 0] / meanVelocity * zoom/zoomDivider), (float)(offsetY + j * zoom - 0.5 * solv.vel[i, j, 1] / meanVelocity * zoom/zoomDivider), (float)(offsetX + i * zoom), (float)(offsetY + j * zoom));
					p.Color = System.Drawing.Color.White;
					gr.DrawLine(p, (float)(offsetX + i * zoom), (float)(offsetY + j * zoom), (float)(offsetX + i * zoom + 0.5 * solv.vel[i, j, 0] / meanVelocity * zoom/zoomDivider), (float)(offsetY + j * zoom + 0.5 * solv.vel[i, j, 1] / meanVelocity * zoom/zoomDivider));
				}
			}
			// Dodaj prepreke
			obstacleVisualization();
		}

		/// <summary>
		/// Interpolacija brzina cetiri susjedna cvora za dobivanje tocnog iznosa brzine u tocki u kojoj se nalazi marker.
		/// </summary>
		/// <param name="marker">Polje markera - imaginarnih cestica koje se pustaju u tok fluida</param>
		/// <param name="i">Indeks i trenutnog markera</param>
		/// <param name="j">Indeks j trenutnog markera</param>
		/// <param name="vel">Polje preko kojega ce biti vracene vrijednosti komponenti brzine</param>
		private void getMarkerVel (double[,,] marker, int i, int j, double[] vel) 
		{
			// Pomocna polja - interpolacije gornjih i donjih susjeda
			double[] A = new double[2];
			double[] B = new double[2];

			// Udaljenost markera od susjeda po x i y smjeru
			double a1, a2, b1, b2;

			// Koordinate donjeg lijevog cvora kvadrata u kojem je trenutni marker
			int ix, iy;

			ix = (int)marker[i, j, 0];
			iy = (int)marker[i, j, 1];

			a1 = marker[i, j, 0] - (int)marker[i, j, 0];
			a2 = 1 - a1;
			b1 = marker[i, j, 1] - (int)marker[i, j, 1];
			b2 = 1 - b1;

			try 
			{
				A[0] = a2 * solv.vel[ix, iy, 0] + a1 * solv.vel[ix + 1, iy, 0];
				A[1] = a2 * solv.vel[ix, iy, 1] + a1 * solv.vel[ix + 1, iy, 1];
				B[0] = a2 * solv.vel[ix, iy + 1, 0] + a1 * solv.vel[ix + 1, iy + 1, 0];
				B[1] = a2 * solv.vel[ix, iy + 1, 1] + a1 * solv.vel[ix + 1, iy + 1, 1];
			}
			catch 
			{
				vel[0] = 0;
				vel[1] = 0;

				return;
			}

			vel[0] = b2 * A[0] + b1 * B[0];
			vel[1] = b2 * A[1] + b1 * B[1];
		}

		/// <summary>
		/// Iscrtavanje rezultata simulacije - vizualizacija strujnica
		/// </summary>
		public void drawStreams () 
		{
			double[] vel = new double[2];
			double[,,] marker = new double[(problemWidth - 1), (problemHeight - 1), 2];
			System.Drawing.Brush b = new System.Drawing.SolidBrush(System.Drawing.Color.FromArgb(63, 184, 184, 184));

			clearPicBox(System.Drawing.Color.Black);

			// Postavi imaginarne cestice jednoliko u sustavu
			for (int i = 1; i < marker.GetLength(0); i++) 
			{
				for (int j = 1; j < marker.GetLength(1); j++) 
				{
					marker[i, j, 0] = i;
					marker[i, j, 1] = j;
				}
			}

			// Za svaki marker izracunaj brzinu, iscrtaj trenutnu poziciju i pomakni ga u ovisnosti o brzini
			for (int i = 0; i < marker.GetLength(0); i++) 
			{
				for (int j = 0; j < marker.GetLength(1); j++) 
				{
					for (int k = 0; k < 100; k++) 
					{
						getMarkerVel(marker, i, j, vel);

						// Ako marker nije izasao iz sustava (kroz istok ili greskom proracuna zbog vremenskog koraka puno veceg od deltaT)
						if (marker[i, j, 0] > 0 && marker[i, j, 0] < problemWidth && marker[i, j, 1] > 0 && marker[i, j, 1] < problemHeight) 
						{
							gr.FillRectangle(b, (float)(offsetX + (marker[i, j, 0]) * zoom), (float)(offsetY + (marker[i, j, 1]) * zoom), (float)2, (float)2);
						}
						marker[i, j, 0] += vel[0] / 3;
						marker[i, j, 1] += vel[1] / 3;
					}
				}
			}
			// Dodaj prepreke
			obstacleVisualization();
		}
	}
}